import os
import shutil
import logging

class AssetFolder:
    def __init__(self, site, path):
        self.path = path
        self.site = site

    def __str__(self):
        return "AssetFolder: [%s]" % self.path

    def copy(self):
        dest_path = os.path.join(self.site.output_path, os.path.basename(self.path))
        logging.debug('Copy Asset Folder from [{}] to [{}]'.format(self.path, dest_path))
        if os.path.isdir(dest_path):
            shutil.rmtree(dest_path)
        shutil.copytree(self.path, dest_path)



import re
import time
import datetime
import calendar

def date_to_rfc822(value, format="%a, %d %b %Y %H:%M:%S +0000"):
    if type(value) == int or type(value) == float:
        return time.strftime(format, time.gmtime(value))
    elif type(value) == str:
        return time.strftime(format, time.strptime(value))
    elif type(value) in [datetime.date, datetime.datetime]:
        return value.strftime(format)
    elif type(value) == time.struct_time:
        return time.strftime(format, value)
    else:
        raise Exception("Unknwon value type!")

def date_to_iso8601(value, format="%Y-%m-%dT%H:%M:%S%z"):
    if type(value) == int or type(value) == float:
        return time.strftime(format, time.gmtime(value))
    elif type(value) == str:
        return time.strftime(format, time.strptime(value))
    elif type(value) in [datetime.date, datetime.datetime]:
        return value.strftime(format)
    elif type(value) == time.struct_time:
        return time.strftime(format, value)
    else:
        raise Exception("Unknwon value type!")

def xml_escape(value):
    if not value:
        return ''
    else:
        XML_CHARS = {
            "<": '&lt;',
            ">": '&gt;',
            "'": '&apos;',
            '"': '&quot;',
            '&': '&amp;',
        }
        def sub(match):
            return XML_CHARS.get(match)
        return re.sub(r"[<>'\"&]", sub, str(value))

def escape_quote(value):
    def sub(match):
        return r'\"'
    if not value:
        return ''
    else:
        return re.sub(r'"', sub, value)



import os
import threading
from http.server import HTTPServer, BaseHTTPRequestHandler, SimpleHTTPRequestHandler

class HttpHandler(SimpleHTTPRequestHandler):
    pass

def serve(site):
    server_address = ('', int(site.cli_options.port) or 0)
    httpd = HTTPServer(server_address, HttpHandler)
    os.chdir(site.output_path)
    print("Listening on {}:{} from {}".format(httpd.server_name, httpd.server_port, os.getcwd()))
    httpd.serve_forever()

def serve_standalone(site):
    """ start and serve in a new thread """
    thread = threading.Thread(target=serve, args=([site]), name="sitekicker-server")
    return thread



import shutil
import os
import logging

class EntryDir:
    def __init__(self, entry, src):
        self.entry = entry
        self.src = src

    def __str__(self):
        return "Entry Dir: %s " % self.src

    def copy(self):
        dest = os.path.join(
            self.entry.output_dir,
            os.path.basename(self.src)
        )
        logging.debug("Copy %s to %s" % (self.src, dest))
        shutil.copytree(self.src, dest)



import os
import yaml
import time
from multiprocessing import Pool, Manager
import collections

from .util import resolve_path, dotdict, get_default_site_options
from .site_tasks import register_site_tasks
from .entry.entry_tasks import register_entry_tasks
from .site_watcher import watch_site
from .site_server import serve

SITE_HOOK_NAMES = [
    'pre-scan',
    'scan',
    'post-scan',
    'pre-mark',
    'mark',
    'post-mark',
    'pre-build',
    'build',
    'post-build',
    'pre-summary',
    'summary',
    'post-summary',
]

ENTRY_HOOK_NAMES = [
    'pre-compile',
    'compile',
    'post-compile',
    'pre-link',
    'link',
    'post-link',
]

class Site:
    def __init__(self, argv_options):
        self.working_path = resolve_path(argv_options.folder)
        # Command line options
        self.cli_options = argv_options
        # Default options
        self.default_options = get_default_site_options()
        # This is the options from sitekicker.yml and command line options
        user_options_dict = dict(self.default_options)
        user_options_dict.update(self.read_sitekicker_yml())
        self.user_options = dotdict(user_options_dict)
        # This is the derived options from user settings, cli options etc
        self.build_options = dotdict({})
        self.build_options.update(self.default_options)
        self.build_options.update(self.user_options)
        self.output_path = resolve_path(argv_options.output_dir or self.user_options.output_dir or '.dist')
        self.build_options.output_path = self.output_path
        self.build_options.working_path = self.working_path
        # data placeholders
        self.time = time.localtime()
        self.timestamp = time.time()
        self.template_registry = []
        self.folders = {}
        self.folder_entries = {}
        self.entries = {}
        self.sorted_entries = []
        self.grouped_entries = {}
        # build hook registry
        self.site_hooks = collections.OrderedDict()
        self.entry_hooks = collections.OrderedDict()
        for hook in SITE_HOOK_NAMES:
            self.site_hooks[hook] = []
        for hook in ENTRY_HOOK_NAMES:
            self.entry_hooks[hook] = []
        # register site task
        register_site_tasks(self)
        register_entry_tasks(self)
        # global task pool for parallel processing
        self.task_pool = Pool()
        self.task_pool_manager = Manager()
        self.snapshots = self.task_pool_manager.dict()

    def reset(self):
        self.time = time.gmtime()
        self.template_registry = []
        self.folders = {}
        self.folder_entries = {}
        self.entries = {}
        self.sorted_entries = []
        self.grouped_entries = {}
        self.task_pool = Pool()
        self.snapshots = self.task_pool_manager.dict()

    def __str__(self):
        return "Site: [%s], output to [%s]" % (self.working_path, self.output_path)

    def read_sitekicker_yml(self):
        site_yml_path = os.path.join(self.working_path, 'sitekicker.yml')
        if os.path.isfile(site_yml_path):
            with open(site_yml_path, 'rt', encoding='utf8') as site_config:
                user_options = dotdict(yaml.load(site_config))
                return user_options
        else:
            raise Exception("sitekicker.yml is not found!")

    def register_site(self, hook, handler):
        if hook in self.site_hooks:
            self.site_hooks[hook].append(handler)
        else:
            raise Exception("Invalid Site Hook name: %s" % hook)

    def register_entry(self, hook, handler):
        if hook in self.entry_hooks:
            self.entry_hooks[hook].append(handler)
        else:
            raise Exception("Invalid Entry Hook name: %s" % hook)

    def build(self):
        """ Default build operation, lazy, only build parts that need to be built """
        self.reset()
        for hook in SITE_HOOK_NAMES:
            if self.site_hooks[hook]:
                for handler in self.site_hooks[hook]:
                    handler(self)

    def watch(self, serve_site=False):
        """ Watch for changes inside the site folder, rebuild items that changed """
        watch_site(self, serve_site)

    def serve(self):
        serve(self)



import jinja2

class EntryTemplate():
    def __init__(self, path, template):
        self.path = path
        self.template = template

    def __str__(self):
        return "Entry Template At: [%s]" % self.path

    def render(self, data):
        return self.template.render(data)



import os
import yaml

from ..util import merge_options

class EnclosureFolder:
    def __init__(self, site, path):
        self.site = site
        self.path = path
        self.options = self.read_folder_options()
        self.combined_options = None

    def __str__(self):
        return "EnclosureFolder: [%s], folder options: %s" % (self.path, self.options)

    def read_folder_options(self):
        folder_yml_path = os.path.join(self.path, 'folder.yml')
        if os.path.isfile(folder_yml_path):
            with open(folder_yml_path, 'r', encoding='utf8') as folder_config:
                return yaml.load(folder_config) or {}
        else:
            return {}

    def get_combined_options(self):
        if self.combined_options is not None:
            return self.combined_options
        else:
            self.combined_options = {}
            merge_options(self.combined_options, self.options)
            enclosing_folder_path = os.path.dirname(self.path)
            enclosing_folder = self.site.folders.get(enclosing_folder_path, None)
            if enclosing_folder:
                enclosing_folder_options = enclosing_folder.get_combined_options()
                merge_options(self.combined_options, enclosing_folder_options)
            return self.combined_options



import os
import logging
import yaml

from ..util import merge_options

class Entry:
    """
    An entry usually a folder with main content file and assets, images etc. It corresponds to a unique url in generated site.
    """
    def __init__(self, site, path):
        self.site = site
        self.path = path
        self.dir = os.path.dirname(path)
        self.read_entry_content()
        self.resolve_entry_options()
        self.id = self.user_options.get('id')
        self.date = self.user_options.get('date')
        self.mtime = os.path.getmtime(self.path)
        perm_link_parts = list(self.options.get('prefix', []))
        perm_link_parts.append(str(self.id))
        self.perm_link = site.user_options['base_url'] + '/' + "/".join(perm_link_parts)
        self.link = '/' + "/".join(perm_link_parts)
        self.external_images = []
        self.linked_images = []
        self.inlined_files = []
        self.linked_files = []
        self.resolve_output_path()

    def __str__(self):
        return "Entry(%s): [%s], %d images, %d external images, %d files, %d inlined files." % (self.id, self.path, len(self.linked_images), len(self.external_images), len(self.linked_files), len(self.inlined_files))

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.date == other.date

    def __gt__(self, other):
        return self.date > other.date

    def __lt__(self, other):
        return self.date < other.date

    def read_entry_content(self):
        with open(self.path, 'rt', encoding='utf8') as mf:
            all_lines = mf.readlines()
            try:
                first_line_index = all_lines.index("---\n", 0)
                second_line_index = all_lines.index("---\n", 1)
                self.user_options = yaml.load(''.join(all_lines[first_line_index+1:second_line_index])) or {}
                self.raw_content = ''.join(all_lines[second_line_index+1:])
            except Exception as e:
                logging.debug("Entry read exception: %s", e)
                self.user_options = {}
                self.raw_content = ''

    def resolve_output_path(self):
        user_output_path = self.options.get('output_path', None)
        if user_output_path is not None:
            self.output_path = os.path.join(self.site.output_path, self.options.get('output_path'))
        else:
            self.output_path = os.path.join(self.site.output_path, *self.options.get('prefix', []), self.id)

    def resolve_entry_options(self):
        self.options = {}
        enclosing_folder = os.path.dirname(self.dir)
        enclosing_folder_options = self.site.folders.get(enclosing_folder).get_combined_options()
        merge_options(self.options, enclosing_folder_options)
        merge_options(self.options, self.user_options)

    def build(self):
        for name, hooks in self.site.entry_hooks.items():
            for hook in hooks:
                hook(self)



import os
import logging
import re
import subprocess
import shlex
import shutil
from ..util import get_image_size
from .entry_file import EntryFile

def compress_resize_image(src, dest, target_width, quality=80, caches=None, full_build=False):
    options_cache_key = "{}-options".format(src)
    options_cache_value = "{},{}".format(target_width, quality)
    cache_hit = (
        src in caches and
        os.path.getmtime(src) == caches[src] and
        dest in caches and
        os.path.isfile(dest) and
        os.path.getmtime(dest) == caches[dest] and
        options_cache_key in caches and
        options_cache_value == caches[options_cache_key]
    )
    if caches is not None and not full_build and cache_hit:
        logging.debug("Cache hit for {}!".format(src))
        return
    logging.debug("IM Resize and Compress image from {} to {}, target_width: {}px, qulity: {}%".format(src, dest, target_width, quality))
    COMPRESS_CMD = "mogrify -write {OUTPUT_PATH} -thumbnail '{OUTPUT_WIDTH}x10000>' -filter Triangle -define filter:support=2 -unsharp 0.25x0.08+8.3+0.045 -dither None -posterize 136 -quality {QUALITY} -define jpeg:fancy-upsampling=off -define png:compression-filter=5 -define png:compression-level=9 -define png:compression-strategy=1 -define png:exclude-chunk=all -interlace none -colorspace sRGB {INPUT_PATH}"
    logging.debug("Copy and compress entry image from %s to %s" % (src, dest))
    subprocess.call(COMPRESS_CMD.format(
        OUTPUT_PATH=shlex.quote(dest),
        QUALITY=quality,
        OUTPUT_WIDTH=target_width,
        INPUT_PATH=shlex.quote(src)),
        shell=True
    )
    if caches is not None:
        caches[src] = os.path.getmtime(src)
        caches[dest] = os.path.getmtime(dest)
        caches[options_cache_key] = options_cache_value

class EntryImage(EntryFile):
    def __init__(self, entry, title, name):
        super().__init__(entry, title, name)
        self.real_width, self.real_height = get_image_size(self.fullpath) if not self.is_external else 0, 0
        matches = re.match(r'(.+)\.([^\.]+)', self.name)
        self.name_no_ext = matches.group(1) if matches else ''
        self.ext = matches.group(2) if matches else ''

    def check_file(self):
        if not self.is_external and not os.path.isfile(self.fullpath):
            self.file_exists=False
            raise Exception("Post [{} > {}] referencing a non-exist local image [{}]!".format(self.entry.id, self.title, self.name))

    def __str__(self):
        return "Entry image: %s" % self.fullpath

    def copy(self):
        logging.debug("Copy entry image from %s to %s" % (self.fullpath, self.dest_fullpath))
        dest_dir = os.path.dirname(self.dest_fullpath)
        if not os.path.isdir(dest_dir):
            os.makedirs(dest_dir)
        if self.entry.site.build_options['compress_image']:
            task_arguments = (
                self.fullpath,
                self.dest_fullpath,
                self.entry.site.build_options['maximum_image_width'],
                self.entry.site.build_options['compress_image_quality'],
                self.entry.site.snapshots,
                self.entry.site.cli_options.full_build
            )
            if not self.entry.site.cli_options.no_parallel:
                self.entry.site.task_pool.apply_async(compress_resize_image, task_arguments)
            else:
                compress_resize_image(*task_arguments)
        else:
            shutil.copy(self.fullpath, self.dest_fullpath)

    def responsive_process(self):
        # generate process task array
        target_widths = self.entry.site.build_options['responsive_image_sizes']
        task_array=[]
        for tw in target_widths:
            if self.real_width<tw:
                save_name=self.name_no_ext + '-' + str(self.real_width) + 'px.' + self.ext
                new_dest_fullpath = os.path.join(os.path.dirname(self.dest_fullpath), save_name)
                logging.debug("Resize(save-no-resize) %s to %s", self.fullpath, new_dest_fullpath)
                arg=(
                    self.fullpath,
                    new_dest_fullpath,
                    self.real_width,
                    self.entry.site.build_options['compress_image_quality'],
                    self.entry.site.snapshots,
                    self.entry.site.cli_options.full_build
                )
            else:
                ratio = self.real_height/self.real_width
                new_height = int(ratio * tw)
                save_name = self.name_no_ext + '-' + str(tw) + 'px.' + self.ext
                new_dest_fullpath = os.path.join(os.path.dirname(self.dest_fullpath), save_name)
                logging.debug("Resize %s to new size: %ix%i, saved to %s", self.fullpath, tw, new_height, new_dest_fullpath)
                arg=(
                    self.fullpath,
                    new_dest_fullpath,
                    tw,
                    self.entry.site.build_options['compress_image_quality'],
                    self.entry.site.snapshots,
                    self.entry.site.cli_options.full_build
                )
            task_array.append(arg)
        # add placeholder image generation task
        task_array.append((
            self.fullpath,
            self.entry.dest_fullpath,
            self.entry.site.build_options['image_placeholder_size'],
            self.entry.site.build_options['image_placeholder_quality'],
            self.entry.site.snapshots,
            self.entry.site.cli_options.full_build
        ))
        # parallel or serial
        if not self.entry.site.cli_options.no_parallel:
            self.entry.site.task_pool.starmap_async(compress_resize_image, task_array)
        else:
            for task in task_array:
                compress_resize_image(*task)



import time
import os
import logging

from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

from .site_server import serve_standalone

class FsChangeHandler(FileSystemEventHandler):
    def __init__(self, site):
        self.site = site

    def on_any_event(self, event):
        if not event.is_directory and event.src_path.endswith('.swp'):
            return
        if event.is_directory and event.event_type=='modified':
            return
        logging.info("Change, DIR: %s, Type: %s, PATH: %s", event.is_directory, event.event_type, event.src_path)
        self.site.build()

def watch_site(site, serve_site=False):
    logging.info("Watching %s...", site)
    ob = Observer()
    serve_thread = None
    change_handler = FsChangeHandler(site)
    for item in os.scandir(site.working_path):
        if item.is_dir() and item.name[0] != '.':
            logging.info("Watching: %s", item.path)
            ob.schedule(change_handler, item.path, recursive=True)
    ob.start()
    if serve_site:
        serve_thread = serve_standalone(site)
        serve_thread.start()
    try:
        while True:
            time.sleep(0.5)
    except KeyboardInterrupt:
        ob.stop()
    ob.join()



import os
import logging
import shutil

class EntryFile:
    """ Represents a binary file inside an entry """
    def __init__(self, entry, title, name):
        self.entry = entry
        self.title = title
        self.name = name
        self.fullpath = os.path.realpath(os.path.join(entry.dir, name))
        self.dest_fullpath = os.path.join(entry.output_path, self.fullpath[len(self.entry.dir)+1:])
        self.is_external=True
        self.file_exists=True
        if os.path.samefile(os.path.commonpath([self.fullpath, entry.dir]), entry.dir):
            self.is_external=False
        self.check_file()

    def check_file(self):
        if not self.is_external and not os.path.isfile(self.fullpath):
            self.file_exists=False
            logging.warning("Post [{} > {}] referencing a non-exist local file: [{}]!".format(self.entry.id, self.title, self.name))

    def __str__(self):
        return "Entry file: %s" % self.fullpath

    def copy(self):
        if not self.file_exists:
            return
        dest_dir = os.path.dirname(self.dest_fullpath)
        logging.debug("Copy entry file from %s to %s" % (self.fullpath, self.dest_fullpath))
        if not os.path.isdir(dest_dir):
            os.makedirs(dest_dir)
        shutil.copyfile(self.fullpath, self.dest_fullpath)



import os
import logging
import jinja2
import json
import time
import collections

from .util import check_is_ignored
from .entry.entry_template import EntryTemplate
from .folder.enclosure_folder import EnclosureFolder
from .folder.template_folder import TemplateFolder
from .folder.asset_folder import AssetFolder
from .folder.entry_folder import EntryFolder

def register_site_tasks(site):
    site.register_site('pre-scan', start_building)
    site.register_site('pre-scan', load_templates)
    site.register_site('pre-scan', prepare_output_path)
    site.register_site('pre-scan', load_previous_build_snapshot)
    site.register_site('scan', scan_site_folders)
    site.register_site('pre-build', convert_entry_folders)
    site.register_site('pre-build', sort_entries_by_date)
    site.register_site('pre-build', group_entries_by_tag)
    site.register_site('build', build_site_entries)
    site.register_site('post-build', end_building)
    site.register_site('pre-summary', dump_build_snapshot)
    site.register_site('pre-summary', copy_assets)
    site.register_site('summary', summary)

def start_building(site):
    print(site)
    site.start_build_time = time.time()

def summary(site):
    site.end_build_time = time.time()
    print("%d seconds used to build!" % int(site.end_build_time - site.start_build_time))

def end_building(site):
    # wait until all tasks are done
    site.task_pool.close()
    site.task_pool.join()

def dump_build_snapshot(site):
    snpashot_path = os.path.join(site.output_path, '.snapshot')
    # manually dump to json to make the output predictable
    with open(snpashot_path, 'wt', encoding='utf8') as sf:
        ordered_data = collections.OrderedDict(sorted(site.snapshots.items(), key=lambda e: e[0]))
        sf.write('{\n')
        for k, v in ordered_data.items():
            sf.write('  {}:{},\n'.format(json.dumps(k), json.dumps(v)))
        sf.write('"":""')
        sf.write('\n}')

def load_previous_build_snapshot(site):
    snpashot_path = os.path.join(site.output_path, '.snapshot')
    if os.path.isfile(snpashot_path):
        with open(snpashot_path, 'rt', encoding="utf8") as sf:
            site.snapshots.clear()
            try:
                data = json.loads(sf.read())
            except:
                data = {}
            site.snapshots.update(data)

def copy_assets(site):
    for path, folder in site.folders.items():
        if isinstance(folder, AssetFolder):
            folder.copy()

def build_site_entries(site):
    print("{} entries found!".format(len(site.entries)))
    for eid, entry in site.entries.items():
        if entry.id and entry.date:
            logging.debug("Building %s", entry)
            entry.build()

def sort_entries_by_date(site):
    valid_entries = [entry for entry in site.entries.values() if entry.id and entry.date]
    site.sorted_entries = sorted(valid_entries, reverse=True)

def group_entries_by_tag(site):
    for entry in site.sorted_entries:
        for t in entry.options.get('tags', []):
            if t not in site.grouped_entries:
                site.grouped_entries[t]=[entry]
            else:
                site.grouped_entries[t].append(entry)

def convert_entry_folders(site):
    for path, folder in site.folders.items():
        if isinstance(folder, EntryFolder):
            folder_entries = folder.find_entries()
            for entry in folder_entries:
                if not entry.id:
                    continue
                if entry.id not in site.entries:
                    site.entries[entry.id] = entry
                else:
                    raise Exception("More than one entry with id: [%s]" % entry.id)
            site.folder_entries[path] = folder_entries

def scan_site_folders(site):
    def detect_folder(path):
        logging.debug("Detecting Folder: [%s]" % path)
        items = os.scandir(path)
        for item in items:
            if item.is_dir() and check_is_ignored(site.build_options['ignore_dirs'], item.path):
                logging.warn("User ignore folder: %s", item.path)
                continue
            if item.is_file() and not item.name == 'folder.yml' and not item.name.startswith('.'):
                logging.debug("Skipping folder file: [%s]", item.path)
            elif not item.is_dir():
                continue
            elif item.name.startswith('.'):
                logging.debug("Skipping Folder: [%s]", item.path)
                continue
            elif EntryFolder.is_entry_folder(item.path):
                site.folders[item.path] = EntryFolder(site, item.path)
                logging.debug("Found New %s", site.folders[item.path])
            else:
                site.folders[item.path] = EnclosureFolder(site, item.path)
                logging.debug("Found New %s", site.folders[item.path])
                detect_folder(item.path)
    site_dirs = os.scandir(site.working_path)
    for sdir in site_dirs:
        if check_is_ignored(site.build_options['ignore_dirs'], sdir.path):
            logging.warn("User ignore folder: %s", sdir.name)
            continue
        if sdir.is_dir() and not sdir.name.startswith('.'):
            if type(site.build_options.asset_dirs)==list and sdir.name in site.build_options.asset_dirs:
                site.folders[sdir.path] = AssetFolder(site, sdir.path)
                logging.debug("Found New %s", site.folders[sdir.path])
            elif sdir.name == site.build_options.template_dir:
                site.folders[sdir.path] = TemplateFolder(site, sdir.path)
                logging.debug("Found New %s", site.folders[sdir.path])
            else:
                site.folders[sdir.path] = EnclosureFolder(site, sdir.path)
                logging.debug("Found New %s", site.folders[sdir.path])
                detect_folder(sdir.path)

def prepare_output_path(site):
    if not os.path.isdir(site.output_path):
        logging.debug("Output path: [%s] not fould, try to create it!", site.output_path)
        os.mkdir(site.output_path)
    elif not os.access(site.output_path, os.W_OK | os.X_OK):
        raise Exception("Output path: [%s] exists, but the permission seems not right, please check!", site.output_path)
    else:
        logging.debug("Output path: [%s] is ok!", site.output_path)


def load_templates(site):
    template_dir = site.build_options['template_dir']
    full_template_path = os.path.realpath(os.path.join(site.working_path, template_dir))
    logging.debug("Scanning template dir: [%s]" % full_template_path)
    site.template_registry = {}
    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(full_template_path),
        trim_blocks=True,
        lstrip_blocks=True,
        extensions=['jinja2.ext.i18n']
    )
    env.install_null_translations()
    # Load Jinja2 Custom Filters
    from .jinja2_filters import xml_escape, date_to_rfc822, date_to_iso8601, escape_quote
    env.filters['xml_escape'] = xml_escape
    env.filters['date_to_rfc822'] = date_to_rfc822
    env.filters['date_to_iso8601'] = date_to_iso8601
    env.filters['escape_quote'] = escape_quote
    template_items = os.scandir(full_template_path)
    for it in template_items:
        if it.is_file() and not it.name.startswith('.') and it.name.endswith('.j2'):
            logging.debug("Find template [%s]@[%s]", it.name, full_template_path)
            site.template_registry[it.name] = EntryTemplate(it.name, env.get_template(it.name))
    logging.debug("%d templates are found!", len(site.template_registry))



import os
import markdown
import logging
import shutil
from datetime import date
from bs4 import BeautifulSoup
import re
import yaml
import html
import jinja2

import sitekicker

from .entry_dir import EntryDir
from .entry_file import EntryFile
from .entry_image import EntryImage
from ..util import get_image_size

def register_entry_tasks(site):
    site.register_entry('pre-compile', resolve_inlined_files)
    site.register_entry('compile', compile_markdown)
    site.register_entry('post-compile', resolve_linked_file)
    site.register_entry('post-compile', resolve_all_images)
    if site.build_options['responsive_images']:
        site.register_entry('post-compile', process_responsive_images_tags)
    site.register_entry('post-compile', setup_meta_tags)
    site.register_entry('link', link_entry)
    site.register_entry('post-link', write_entry_output)
    site.register_entry('post-link', copy_entry_files)
    if site.build_options['responsive_images']:
        site.register_entry('post-link', process_responsive_entry_images)
    else:
        site.register_entry('post-link', copy_entry_images)
    site.register_entry('post-link', summary)

def summary(entry):
    print(entry)

def copy_entry_images(entry):
    for img in entry.linked_images:
        if not img.is_external:
            img.copy()

def copy_entry_files(entry):
    for fi in entry.linked_files:
        if not fi.is_external:
            fi.copy()

def process_responsive_entry_images(entry):
    for img in entry.linked_images:
        img.responsive_process()

def resolve_inlined_files(entry):
    """ Inlined files must be read and insert to its anchor point, before compilation """
    entry.inlined_files = []
    def read_file(match):
        name = match.group(1)
        fullpath = os.path.join(entry.dir, name)
        entry.inlined_files.append(fullpath)
        logging.debug("Find inlined file: [%s]", fullpath)
        with open(fullpath, 'rt', encoding="utf8") as f:
            return f.read()
    raw_content = entry.raw_content
    # read and insert file content for [[file: filename]] tags before build
    entry.raw_content = re.sub(r'\[\[file: ([^]]+)\]\]', read_file, raw_content)

def compile_markdown(entry):
    """ Build the main content of the entry, which is written in markdown at the moment """
    md_out = markdown.markdown(
        entry.raw_content,
        extensions=[
            'markdown.extensions.extra',
            'pymdownx.tilde',
            'pymdownx.magiclink',
            'pymdownx.arithmatex',
        ],
        output_format="html5"
    )
    entry.compile_output = md_out

def setup_meta_tags(entry):
    entry.options['meta_tags'] = set()
    # Detect MathJax usage
    if re.search(r'\\\(.*\\\)|\\\[.*\\\]', entry.compile_output):
        entry.options['meta_tags'].add('math')
    # Detect Images
    if entry.external_images or entry.linked_images:
        entry.options['meta_tags'].add('image')
        if re.search(r'srcset', entry.compile_output):
            entry.options['meta_tags'].add('responsive-image')
        if re.search(r'lazyload', entry.compile_output):
            entry.options['meta_tags'].add('lazyload-image')
    # Detect Code Blocks
    if re.search(r'\<code[^>]*>', entry.compile_output) or re.search(r'\<pre[^>]*>', entry.compile_output):
        entry.options['meta_tags'].add('code')

def resolve_linked_file(entry):
    """ Find out all files linked with <a> """
    entry.linked_files = []
    def tap_link(link_match):
        link_text = link_match.group(0)
        link = BeautifulSoup(link_text, 'html.parser').a
        attrs = link.attrs
        href = attrs.get('href')
        title = link.string
        if href.startswith('http://') or href.startswith('https://') or href.startswith('//'):
            pass
        else:
            entry.linked_files.append(EntryFile(entry, title, href))
    re.sub(r'\<a\s+[^>]*\s*\>([^<]+)</a>', tap_link, entry.compile_output)

def resolve_all_images(entry):
    """ Find out all images used by <image/> tag """
    entry.external_images = []
    entry.linked_images = []
    def sub_img(image_match):
        img_text = image_match.group(0)
        img = BeautifulSoup(img_text, 'html.parser').img
        attrs = img.attrs
        title = attrs.get('alt') or ''
        name = attrs.get('src') or attrs.get('data-src')
        if name.startswith('http://') or name.startswith('https://') or name.startswith('//'):
            entry.external_images.append(name)
        else:
            entry.linked_images.append(EntryImage(entry, title, name))
    re.sub(r'\<img\s+[^>]*\s*\>', sub_img, entry.compile_output)

def process_responsive_images_tags(entry):
    # Substitute <img> tag for lazy-responsive-load
    def sub_img(image_match):
        img_text = image_match.group(0)
        img = BeautifulSoup(img_text, 'html.parser').img
        attrs = img.attrs
        src = attrs.get('src') or attrs.get('data-src')
        alt = html.escape(attrs.get('alt') or src)
        classes = attrs.get('class') or []
        if src.startswith('http://') or src.startswith('https://') or src.startswith('//'):
            sub = '<img data-src="' + src + '" class="lazyload" />'
        else:
            alt = attrs.get('alt', '')
            src_parts = src.rsplit(sep='.', maxsplit=1)
            src_width, src_height = get_image_size(os.path.join(entry.dir, src))
            srcsets = []
            for width in entry.site.build_options['responsive_image_sizes']:
                if src_width<width:
                    srcsets.append(src_parts[0] + '-'+ str(src_width) +'px' + '.' + src_parts[1] + ' ' + str(src_width) + 'w')
                    break
                else:
                    srcsets.append(src_parts[0] + '-'+ str(width) +'px' + '.' + src_parts[1] + ' ' + str(width) + 'w')
            srcset_text = ','.join(srcsets)
            default_src = src_parts[0] + '-'+ str(entry.site.build_options['responsive_image_sizes'][1]) +'px' + '.' + src_parts[1]
            lqip_src = src_parts[0] + '-'+ str(entry.site.build_options['image_placeholder_size']) +'px' + '.' + src_parts[1]
            sub = '<img style="max-width: {max_width}px; max-height: {max_height}px;" src="{lqip_src}" data-src="{default_src}" data-sizes="auto" data-srcset="{src_set}" class="lazyload lqip-blur {classes}" alt="{alt}"/>'.format(max_width=str(src_width), max_height=str(src_height), lqip_src=lqip_src, default_src=default_src, src_set=srcset_text, alt=alt, classes=' '.join(classes))
        return sub
    entry.compile_output = re.sub(r'\<img\s+[^>]*\s*\>', sub_img, entry.compile_output)

def link_entry(entry):
    """ Link entry content with header, footer to get the final html output for the entry """
    entry_template_name = entry.options.get('layout', 'default')+'.j2'
    page_template = entry.site.template_registry.get(entry_template_name)
    if not page_template:
        if entry_template_name=='default.j2':
            page_template=jinja2.Template('{{ content }}')
        else:
            raise Exception("Entry[{}] using invalid template: [{}]".format(entry.id, entry_template_name))
    template_data = {
        'sitekicker': sitekicker,
        'site': entry.site,
        'entry': entry,
        'entry_content': entry.compile_output,
        'perm_link': entry.perm_link,
    }
    template_data.update(entry.options)
    entry.html_output = page_template.render(template_data)

def write_entry_output(entry):
    """ Save the final html output to file """
    logging.debug('Output path: %s', entry.output_path)
    if not os.path.isdir(entry.output_path):
        os.makedirs(entry.output_path)
    output_name = entry.options.get('output_name', 'index.html')
    output_path = os.path.join(entry.output_path, output_name)
    with open(output_path, 'wt', encoding="utf8") as of:
        logging.debug("Writing built html to: %s", output_path)
        of.write(entry.html_output)



import os
import argparse
import sitekicker
import subprocess
import re

def resolve_path(path):
    path = os.path.expanduser(path)
    return os.path.abspath(path)

def check_is_ignored(patterns, path):
    if not patterns:
        return False
    for p in patterns:
        if re.search(p, path) is not None:
            return True
    return False

def remove_list_duplicate(lst):
    """ Remove duplicate from a list, maintain original order """
    seen = set()
    seen_add = seen.add
    return [x for x in lst if not (x in seen or seen_add(x))]

def merge_options(to_options, from_options):
    for k, v in from_options.items():
        if k == 'tags':
            if not k in to_options:
                to_options[k] = list()
            to_options[k].extend(v)
            to_options[k]=remove_list_duplicate(to_options[k])
        elif k == 'prefix':
            if not k in to_options:
                to_options[k] = list()
            if type(v)==list:
                to_options[k].extend(v)
            else:
                to_options[k].append(v)
        else:
            to_options[k]=v

def parse_command_line_options(arguments):
    ap = argparse.ArgumentParser(description="SiteKicker")
    ap.add_argument(
        "--log-level",
        default='info',
        choices=['info', 'debug', 'warning', 'error', 'critical'],
        dest="log_level",
        help="Set log level, default is warning"
    )
    ap.add_argument('--no-parallel', action="store_true", default=False, help="Do not use parallel image processing, this will make build slower, default is False")
    ap.add_argument('--serve', '-s', action="store_true", default=False, help="Serve the built contents with a local server for preview, default is False")
    ap.add_argument('--watch', '-w', action="store_true", default=False, help="Watch for changes and rebuild, default is False")
    ap.add_argument('--full-build', '-f', action="store_true", default=False, help="Build everything from scratch, ignore all caches, it would slow down the build, default is False")
    ap.add_argument('--version', '-V', action="version", version=sitekicker.version, help="Show version number")
    ap.add_argument('--port', '-p', default="8000", help="Default port for the local preview server to listen")
    ap.add_argument('--output-dir', '-o', default="", help="Directory to write build output")
    ap.add_argument("folder", nargs="?", default=".", help="Folder to process, default is current directory")
    argv_options = ap.parse_args(arguments)
    return argv_options

def get_default_site_options():
    return dotdict({
        'name': "An Awesome Website",
        'base_url': '',
        'absolute_urls': False,
        'template_dir': 'templates',
        'output_dir': '.dist',
        'asset_dirs': ['assets'],
        'content_dirs': [],
        'ignore_dirs': [],
        'copy_hidden': False,
        'responsive_images': False,
        'responsive_image_sizes': [500, 1000, 1500],
        'image_placeholder_size': 48,
        'image_placeholder_quality': 15,
        'maximum_image_width': '1500',
        'compress_image': True,
        'compress_image_quality': 80
    })

class dotdict(dict):
    def __init__(self, *args, **kwargs):
        super(dotdict, self).__init__(*args, **kwargs)
        self._set_from_args(args, kwargs)

    def _set_from_args(self, args, kwargs):
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    if isinstance(v, dict):
                        self[k] = dotdict(v)
                    else:
                        self[k] = v
        if kwargs:
            for k, v in kwargs.items():
                if isinstance(v, dict):
                    self[k] = dotdict(v)
                else:
                    self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(dotdict, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(dotdict, self).__delitem__(key)
        del self.__dict__[key]

    def update(self, *args, **kwargs):
        super(dotdict, self).update(*args, **kwargs)
        self._set_from_args(args, kwargs)

def get_image_size(src):
    im_raw_size = subprocess.check_output(['identify', src])
    im_raw_size_match = re.search(r'\s(\d+)x(\d+)\s', im_raw_size.decode('utf8') if im_raw_size else '')
    if im_raw_size_match:
        return int(im_raw_size_match.group(1)), int(im_raw_size_match.group(2))
    else:
        return 0, 0



import os
import sys
import logging
import time
import subprocess

from .site import Site
from .util import parse_command_line_options

def main():
    argv_options=parse_command_line_options(sys.argv[1:])
    # Logging Configuration
    logging_level = getattr(logging, argv_options.log_level.upper(), logging.WARNING)
    logging.basicConfig(
        level=logging_level,
        format="%(asctime)s::%(levelname)s::%(message)s",
        datefmt="%H:%M:%S"
    )
    site = Site(argv_options)
    logging.info("%s", site)

    if argv_options.watch:
        site.build()
        if argv_options.serve:
            site.watch(True)
        else:
            site.watch()
    else:
        site.build()
        if argv_options.serve:
            try:
                site.serve()
            except KeyboardInterrupt:
                print('exiting')
            except e:
                print(e)

if __name__ == '__main__':
    try:
        main()
    except Exception as e:
        print(e)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['sitekicker', 'sitekicker.entry', 'sitekicker.folder']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4',
 'jinja2',
 'markdown',
 'pygments',
 'pymdown-extensions',
 'pyyaml',
 'watchdog']

setup_kwargs = {
    'name': 'sitekicker',
    'version': '1.0.3',
    'description': 'Another static site generator',
    'long_description': None,
    'author': 'Minghao Ni',
    'author_email': 'niminghao804@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nmhnmh/sitekicker',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



import os
import glob

from ..entry.entry import Entry

class EntryFolder:
    def __init__(self, site, path):
        self.path = path
        self.site = site

    def __str__(self):
        return "EntryFolder: [%s]" % self.path

    def find_entries(self):
        entries = []
        glob_pattern = os.path.join(self.path, '*.md')
        main_candidates = glob.glob(glob_pattern)
        if(main_candidates):
            for candidate in main_candidates:
                if EntryFolder.detect_file_content(candidate):
                    entries.append(Entry(self.site, candidate))
        return entries

    @staticmethod
    def is_entry_folder(path):
        """ Detect if a folder contains a valid entry inside it """
        glob_pattern = os.path.join(path, '*.md')
        main_candidates = glob.glob(glob_pattern)
        if(main_candidates):
            for candidate in main_candidates:
                return EntryFolder.detect_file_content(candidate)
            else:
                return False
        else:
            False

    @staticmethod
    def detect_file_content(path):
        """ Check if a file is valid entry file, by check its front matter """
        with open(path, 'rt', encoding='utf8') as mf:
            all_lines = ''.join(mf.readlines())
            try:
                # valid entry must have valid "front matter" block(options block)
                start_index = all_lines.index("---\n", 0)
                end_index = all_lines.index("---\n", 4)
                # valid entry must have id, title
                all_lines.index("\nid: ", start_index, end_index)
                all_lines.index("\ntitle: ", start_index, end_index)
                # all_lines.index("\ndate: ", start_index, end_index)
            except Exception:
                return False
            else:
                return True
